@@ -258,7 +258,8 @@ class Session(Workspace):
 
         :param query: cypher query
         :param parameters: dictionary of parameters
-        :param kwargs: additional keyword parameters
+        :param kwargs: additional keyword parameters.
+            These take precedence over parameters passed as ``parameters``.
 
         :raises SessionError: if the session has been closed.
 
@@ -286,10 +287,11 @@ class Session(Workspace):
             self._result_error
         )
         bookmarks = self._get_all_bookmarks()
+        parameters = dict(parameters or {}, **kwargs)
         self._auto_result._run(
             query, parameters, self._config.database,
             self._config.impersonated_user, self._config.default_access_mode,
-            bookmarks, **kwargs
+            bookmarks
         )
 
         return self._auto_result
@@ -410,7 +412,7 @@ class Session(Workspace):
             At most one transaction may exist in a session at any point in time.
             To maintain multiple concurrent transactions, use multiple concurrent sessions.
 
-            Note: For auto-transaction (Session.run) this will trigger an consume for the current result.
+            Note: For auto-transaction (Session.run) this will trigger a consume for the current result.
 
         :param metadata:
             a dictionary with metadata.
